# Generated by Django 3.2.19 on 2023-06-05 07:22

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import resolwe.flow.models.fields
from resolwe.flow.models.fields import VersionField


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("flow", "0008_annotations"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationpreset",
            name="contributor",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                to="auth.user",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="created",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="modified",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="slug",
            field=resolwe.flow.models.fields.ResolweSlugField(
                max_length=100, populate_from="name", unique_with=("version",)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationpreset",
            name="version",
            field=VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="annotationpreset",
            name="name",
            field=models.CharField(max_length=100),
        ),
    ]
