# Generated by Django 3.2.12 on 2022-08-22 13:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0005_duplicate_data_dependency"),
    ]

    operations = [
        migrations.AddField(
            model_name="relation",
            name="descriptor",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="relation",
            name="descriptor_dirty",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="relation",
            name="descriptor_schema",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="flow.descriptorschema",
            ),
        ),
    ]
