# Generated by Django 5.1.5 on 2025-02-12 06:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0035_move_version_field_to_resolwe"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationvalue",
            name="deleted",
            field=models.BooleanField(default=False),
        ),
    ]
